use fs::OpenOptions;
use std::io::{Read, Write};
use std::net::{SocketAddr, TcpListener};
use std::process::exit;
use std::thread::{park, sleep, spawn};
use std::time::Duration;
use std::{fs, panic};
use sysinfo::{CpuRefreshKind, MINIMUM_CPU_UPDATE_INTERVAL, RefreshKind, System};

fn get_usage(s: &mut System) -> u8 {
    s.refresh_cpu_usage();
    s.cpus()
        .iter()
        .map(|item| item.cpu_usage() as u8)
        .max_by(|a, b| a.partial_cmp(b).unwrap())
        .unwrap()
}
static ACTIONS: &[&str] = &["ondemand", "powersave"];
fn set_governor(cpu_count: u16, target: &str) -> bool {
    if !ACTIONS.contains(&target) {
        eprintln!("asked to set governor to {}, rejecting!", target);
        return false;
    }
    println!("setting governor to {}", target);
    for i in 0..cpu_count {
        let path = format!("/sys/devices/system/cpu/cpu{}/cpufreq/scaling_governor", i);
        let mut file = OpenOptions::new().write(true).open(&path).unwrap();
        file.write_all(target.as_bytes()).unwrap();
    }
    true
}
fn watcher(cpu_count: u16, mut s: System) {
    let upscale_reset_duration = Duration::from_secs(5);
    let downscale_reset_duration = Duration::from_secs(3);
    let refresh_interval = Duration::from_millis(500);
    let upscale_percentage = 90;
    let downscale_percentage = 50;

    if refresh_interval < MINIMUM_CPU_UPDATE_INTERVAL {
        panic!(
            "refresh interval too short, should be minimum {:?}",
            MINIMUM_CPU_UPDATE_INTERVAL
        )
    }

    let upscale_reset_amount = upscale_reset_duration.div_duration_f32(refresh_interval) as u8;
    let downscale_reset_amount = downscale_reset_duration.div_duration_f32(refresh_interval) as u8;
    let mut upscale_remaining = upscale_reset_amount;
    let mut downscale_remaining = downscale_reset_amount;

    sleep(MINIMUM_CPU_UPDATE_INTERVAL); //To be sure next update is good.
    loop {
        let max = get_usage(&mut s);
        if max > upscale_percentage {
            upscale_remaining -= 1;
            if upscale_remaining == 0 {
                set_governor(cpu_count, "ondemand");
                upscale_remaining = upscale_reset_amount;
            }
        } else {
            upscale_remaining = upscale_reset_amount;
        }
        if max < downscale_percentage {
            downscale_remaining -= 1;
            if downscale_remaining == 0 {
                set_governor(cpu_count, "powersave");
                downscale_remaining = downscale_reset_amount;
            }
        } else {
            downscale_remaining = downscale_reset_amount;
        }
        sleep(refresh_interval);
    }
}

fn buffer_to_action(buf: &[u8]) -> Option<String> {
    let string = String::from_utf8_lossy(buf);
    let mut data = string.split(" ");
    data.next();
    let str = data.next()?.strip_prefix("/")?;
    Some(str.to_owned())
}

fn server(cpu_count: u16) {
    let addr = SocketAddr::from(([0, 0, 0, 0], 17806));
    let listener = TcpListener::bind(addr).unwrap();
    for stream_result in listener.incoming() {
        let mut stream = stream_result.unwrap();
        let buf = &mut [0; 128];
        stream.read(buf).unwrap();
        let optional = buffer_to_action(buf);
        if optional.is_none() {
            eprintln!("Failed to parse buffer");
            continue;
        }
        let result = set_governor(cpu_count, &*optional.unwrap());
        let status = if result { "200 OK" } else { "400 Bad Request" };
        stream.write("HTTP/1.1 ".as_bytes()).unwrap();
        stream.write(status.as_bytes()).unwrap();
        stream
            .write("\r\nContent-Length: 0\r\n\r\n".as_bytes())
            .unwrap();
    }
}

fn main() {
    //Thread crash handler
    let orig_hook = panic::take_hook();
    panic::set_hook(Box::new(move |panic_info| {
        // invoke the default handler and exit the process
        orig_hook(panic_info);
        println!("panic hook, exiting autoPowersave!");
        exit(1);
    }));

    let s = System::new_with_specifics(
        RefreshKind::nothing().with_cpu(CpuRefreshKind::nothing().with_cpu_usage()),
    );
    let cpu_count = s.cpus().len() as u16;
    spawn(move || watcher(cpu_count, s));
    spawn(move || server(cpu_count));
    park();
}
